'''模型最终训练'''
import random
from collections import OrderedDict

import torch
from torch import nn, optim
import numpy as np
from scipy.io import loadmat, savemat
from torch.utils.data import DataLoader, random_split
import os
import argparse
from sklearn.metrics import cohen_kappa_score, accuracy_score, recall_score
from torch.utils.tensorboard import SummaryWriter
import itertools
from os import makedirs, path

from datasets.HLDataset import HLDatasetV1
from model.pt import PT
from model.ft import FT
from model.encoder_hl import Encoder_hl
from utils import train_fusion, test_fusion




ROOT = None
RUN = 10
SAMPLE_PER_CLASS = [10]
LR = 1e-3
seed = 84
torch.manual_seed(seed)
torch.manual_seed(seed)
torch.cuda.manual_seed(seed)
torch.cuda.manual_seed_all(seed)
random.seed(seed)
np.random.seed(seed)
import os
os.environ['CUDA_VISIBLE_DEVICES'] = '2'
DEVICE = 'cuda' if torch.cuda.is_available() else torch.device('cpu')
MODEL_NAME = 'CCL_FT'

def main(datasetName, n_sample_per_class, run):
    logger = SummaryWriter(ROOT)

    data = loadmat("./Data/hyper_lidar_{}.mat".format(datasetName))['data']
    label = loadmat("./Data/true_label_{}.mat".format(datasetName))['AllTrueLabel']
    trainLabel = loadmat("./trainTestSplit/{}/sample{}_run{}.mat".format(datasetName, n_sample_per_class, run))['train_gt']
    testLabel = loadmat("./trainTestSplit/{}/sample{}_run{}.mat".format(datasetName, n_sample_per_class, run))['test_gt']
    # trainLabel = loadmat("./Data/nobg/houston/sample10_run0.mat")['train_gt']
    # testLabel = loadmat("./Data/nobg/houston/sample10_run0.mat")['test_gt']

    # 数据转换
    data, trainLabel, testLabel = data.astype(np.float32), trainLabel.astype(np.int), testLabel.astype(np.int)
    bands = data.shape[2]
    hsi_bands = bands - 1
    label = label.astype(np.int)
    nc = int(np.max(trainLabel))


    train_dataset = HLDatasetV1(data, trainLabel, patchsz=args.img_dim)
    val_dataset = HLDatasetV1(data, testLabel, patchsz=args.img_dim)
    train_loader = DataLoader(dataset=train_dataset,
                              batch_size=16,
                              shuffle=True,
                              num_workers=args.num_workers,
                              pin_memory=True,
                              drop_last=True)
    val_dataloader = DataLoader(dataset=val_dataset,
                                batch_size=128,
                                shuffle=False,
                                num_workers=args.num_workers,
                                pin_memory=True,
                                drop_last=True)

    model = PT(args, hsi_bands)
    pre_path = args.pre_path
    checkpoint = torch.load(pre_path)
    state_dict = checkpoint['state_dict']
    model.load_state_dict(state_dict)
    # params = list(model.named_parameters())
    # print(params)

    # translator = model.translator
    encoder_h = model.backbone_hsi
    encoder_l = model.backbone_lidar
    model_ft = FT(nc)



    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model_ft.parameters(), lr=LR)
    # optimizer = optim.Adam(itertools.chain(resnet_h.parameters(), resnet_l.parameters(), model_fusion.parameters()), lr=LR)


    res = torch.zeros((3, EPOCHS))
    res_xi = torch.zeros(3)
    best_acc, best_AA, best_OA, best_Kappa = 0, 0, 0, 0
    for epoch in range(EPOCHS):
        print('*'*5 + 'Epoch:{}'.format(epoch) + '*'*5)
        # hsi
        trainLoss = train_fusion(encoder_h, encoder_l, model_ft, criterion=criterion,
                                 optimizer=optimizer, dataLoader=train_loader, device=DEVICE)
        logger.add_scalar('Loss/train', trainLoss, epoch)
        acc, evalLoss, AA, OA, Ka = test_fusion(encoder_h, encoder_l, model_ft,
                                                criterion=criterion, dataLoader=val_dataloader, device=DEVICE)
        if best_acc < acc:
            best_acc = acc
            best_AA = AA
            best_OA = OA
            best_Kappa = Ka
            torch.save(encoder_h.state_dict(), os.path.join(ROOT, 'encoder_h_best.pkl'))
            torch.save(encoder_l.state_dict(), os.path.join(ROOT, 'encoder_l_best.pkl'))
            torch.save(model_ft.state_dict(), os.path.join(ROOT, 'model_ft_best.pkl'))
        logger.add_scalar('Loss/test', evalLoss, epoch)
        logger.add_scalar('acc/test', acc, epoch)
        print('epoch:{} trainLoss:{:.8f} evalLoss:{:.8f} acc:{:.4f}'.format(epoch, trainLoss, evalLoss, acc))
        print('AA:{:.4f} OA:{:.4f} Kappa:{:.4f}'.format(AA, OA, Ka))
        print('best result---acc:{:.4f} AA:{:.4f} OA:{:.4f} Kappa:{:.4f}'.format(best_acc, best_AA, best_OA, best_Kappa))
        print('*'*18)
        res[0][epoch], res[1][epoch], res[2][epoch] = trainLoss, evalLoss, acc
        if epoch % 10 == 0:
            torch.save(model_ft.state_dict(),
                       os.path.join(ROOT, '{}_sample{}_run{}_epoch{}.pkl'.format(datasetName, n_sample_per_class,
                                                                                 run, epoch)))
    res_xi[0], res_xi[1], res_xi[2] = best_AA, best_OA, best_Kappa
    tmp_simsiam = res.numpy()
    tmp_xi_simsiam = res_xi.numpy()
    savemat(os.path.join(ROOT, 'res.mat'), {'trainLoss': tmp_simsiam[0], 'evalLoss': tmp_simsiam[1], 'acc': tmp_simsiam[2]})
    savemat(os.path.join(ROOT, 'res_xi.mat'), {'best_AA': tmp_xi_simsiam[0], 'best_OA': tmp_xi_simsiam[1], 'best_Kappa': tmp_xi_simsiam[2]})
    return res, res_xi


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='CCL_FT')
    parser.add_argument('--name', type=str, default='trento',
                        help='The name of dataset')
    parser.add_argument('--epoch', type=int, default=300,
                        help='模型的训练次数')
    parser.add_argument('--lr', type=float, default=1e-3,
                        help='learning rate')
    parser.add_argument('--num_workers', type=int, default=8, help='num of workers to use')
    parser.add_argument('--img_dim', default=9, type=int)
    parser.add_argument('--arch', default='resnet18', help='model name is used for training')

    parser.add_argument('--feat_dim', default=1024, type=int, help='feature dimension')
    parser.add_argument('--num_proj_layers', type=int, default=2, help='number of projection layer')
    parser.add_argument('--exp_dir', type=str, default=None, help='path to experiment directory')
    parser.add_argument('--trial', type=str, default=None, help='trial id')
    parser.add_argument('--pre_path', type=str,
                        default='/home3/zx/CCL-master/trento/ccl_pt_01/ckpt_epoch_9_ccl_pt_01.pth',
                        help='pre hsi lidar 1是180')
    parser.add_argument('--n', default='hsi_lidar', help='pre is hsi or hsi_lidar')
    args = parser.parse_args()
    EPOCHS = args.epoch
    datasetName = args.name
    LR = args.lr
    print('*'*5 + datasetName + '*'*5)
    for i, num in enumerate(SAMPLE_PER_CLASS):
        print('*' * 5 + 'SAMPLE_PER_CLASS:{}'.format(num) + '*' * 5)
        res = torch.zeros((RUN, 3, EPOCHS))
        res_xi = torch.zeros((RUN, 3))
        for r in range(RUN):
            print('*' * 5 + 'run:{}'.format(r) + '*' * 5)
            ROOT = '{}/ccl_ft/{}/{}'.format(datasetName, num, r)
            if not os.path.isdir(ROOT):
                os.makedirs(ROOT)
            res[r], res_xi[r] = main(datasetName, num, r)
            # res_xi[r][0] = torch.from_numpy(
            #     loadmat('{}/ccl_ft/{}/{}/res_xi.mat'.format(datasetName, num, r))['best_AA'])
            # res_xi[r][1] = torch.from_numpy(
            #     loadmat('{}/ccl_ft/{}/{}/res_xi.mat'.format(datasetName, num, r))['best_OA'])
            # print(res_xi[r][1])
            # res_xi[r][2] = torch.from_numpy(
            #     loadmat('{}/ccl_ft/{}/{}/res_xi.mat'.format(datasetName, num, r))['best_Kappa'])
        mean = torch.mean(res, dim=0)  # [3, EPOCHS]
        mean_xi = torch.mean(res_xi, dim=0)
        print('mean result---AA:{:.4f} OA:{:.4f} Kappa:{:.4f}'.format(mean_xi[0], mean_xi[1], mean_xi[2]))
        tmp = res.numpy()
        tmp_xi = res_xi.numpy()
        savemat(os.path.join('{}/ccl_ft/{}'.format(datasetName, num), 'res.mat'),
                {'trainLoss': tmp[0], 'evalLoss': tmp[1], 'acc': tmp[2]})
        savemat(os.path.join('{}/ccl_ft/{}'.format(datasetName, num), 'res_xi.mat'),
                {'best_AA': tmp_xi[0], 'best_OA': tmp_xi[1], 'best_Kappa': tmp_xi[2]})
    print('*'*5 + 'FINISH' + '*'*5)

# simsiam_hl_1 houston 没加BN 50轮 2加bn 3反向传播 614减少两层



