import torch
from torch.utils.data import Dataset
import numpy as np

class HLDataset(Dataset):
    def __init__(self, data, label, patchsz=1, transform=None):
        '''
        :param data: [h, w, bands]
        :param label: [h, w]
        :param n_components: scale
        :param patchsz: scale
        '''
        super(HLDataset, self).__init__()
        self.data = data # [h, w, bands]
        self.h, self.w = self.data.shape[0], self.data.shape[1]
        bands = self.data.shape[2]

        self.data_hsi = self.data[:, :, :bands - 1]
        self.data_lidar = self.data[:, :, bands -1]
        [m, n] = self.data_lidar.shape
        self.data_lidar = self.data_lidar.reshape((m, n, 1))
        self.label = label # [h, w]
        self.patchsz = patchsz


        self.data_hsi = self.Normalize(self.data_hsi)
        self.data_hsi = self.addMirror(self.data_hsi)
        self.data_lidar = self.Normalize(self.data_lidar)
        self.data_lidar = self.addMirror(self.data_lidar)



    # 数据归一化
    def Normalize(self, data):
        h, w, c = data.shape
        data = data.reshape((h * w, c))
        data -= np.min(data, axis=0)
        data /= np.max(data, axis=0)
        data = data.reshape((h, w, c))
        return data

    # 添加镜像
    def addMirror(self, data):
        dx = self.patchsz // 2
        h, w, bands = data.shape
        mirror = None
        if dx != 0:
            mirror = np.zeros((h + 2 * dx, w + 2 * dx, bands))
            mirror[dx:-dx, dx:-dx, :] = data
            for i in range(dx):
                # 填充左上部分镜像
                mirror[:, i, :] = mirror[:, 2 * dx - i, :]
                mirror[i, :, :] = mirror[2 * dx - i, :, :]
                # 填充右下部分镜像
                mirror[:, -i - 1, :] = mirror[:, -(2 * dx - i) - 1, :]
                mirror[-i - 1, :, :] = mirror[-(2 * dx - i) - 1, :, :]
        return mirror


    def __len__(self):
        return self.h * self. w

    def __getitem__(self, index):
        '''
        :param index:
        :return: 元素光谱信息， 元素的空间信息， 标签
        '''
        l = index // self.w  # 得到行列坐标
        c = index % self.w
        neighbor_hsi = self.data_hsi[l:l + self.patchsz, c:c + self.patchsz, :]
        x1 = torch.tensor(neighbor_hsi, dtype=torch.float32)
        neighbor_lidar = self.data_lidar[l:l + self.patchsz, c:c + self.patchsz, :]
        x2 = torch.tensor(neighbor_lidar, dtype=torch.float32)
        label = self.label[l, c] - 1
        return [x1, x2], torch.tensor(label, dtype=torch.long)

class HLDatasetV1(HLDataset):
    def __init__(self, data, label, patchsz=1):
        super().__init__(data, label, patchsz)
        self.sampleIndex = list(zip(*np.nonzero(self.label)))

    def __len__(self):
        return len(self.sampleIndex)

    def __getitem__(self, index):
        l, c = self.sampleIndex[index]
        neighbor_hsi = self.data_hsi[l:l + self.patchsz, c:c + self.patchsz, :]
        x1 = torch.tensor(neighbor_hsi, dtype=torch.float32)
        neighbor_lidar = self.data_lidar[l:l + self.patchsz, c:c + self.patchsz, :]
        x2 = torch.tensor(neighbor_lidar, dtype=torch.float32)
        label = self.label[l, c] - 1
        return [x1, x2], torch.tensor(label, dtype=torch.long)
