import torch
from torch.nn import init
from torch import nn
import os
from scipy.io import loadmat
import random
from sklearn.metrics import cohen_kappa_score, accuracy_score, recall_score
import numpy as np
import math
CNAME={
    # 'c0': '#ffffff',
    # 'c1': '#DFBEA7',
    # 'c2': '#8866CC',
    # 'c3': '#B1D3EC',
    # 'c4': '#AA6639',
    # 'c5': '#2D882D',
    # 'c6': '#FFFF00'}

    # 'c1': '#6E8B3D',
    # 'c2': '#339933',
    # 'c3': '#71C671',
    # 'c4': '#006633',
    # 'c5': '#996600',
    # 'c6': '#B1D3EC',
    # 'c7': '#0033CC',
    # 'c8': '#8866CC',
    # 'c9': '#FFFF00',
    # 'c10': '#ffcccc',
    # 'c11': '#FFC080',
    # 'c12': '#CDAA7D',
    # 'c13': '#CD950C',
    # 'c14': '#8B3E2F',
    # 'c15': '#A52A2A'}

    'c1': '#2D882D',
    'c2': '#73E673',
    'c3': '#B6F2B6',
    'c4': '#AA6639',
    'c5': '#DFBEA7',
    'c6': '#B1D3EC',
    'c7': '#4a86e8',
    'c8': '#2662c5',
    'c9': '#FFFF00',
    'c10': '#d91818',
    'c11': '#0033CC'}
COLOR = list(zip(*list(CNAME.items())))[-1]

# 颜色十六进制数转换为RGB
def hex2rgb(hex):
    rgb = []
    for i in range(1, len(hex), 2):
        tmp = '0x' + hex[i:i+2]
        rgb.append(eval(tmp))
    return np.array(rgb, dtype=np.uint8)


def plot_map(gt):
    map = np.zeros_like(gt, dtype=np.uint8)
    map = np.expand_dims(map, axis=-1)
    map = np.repeat(map, 3, axis=-1)
    gt = gt.astype(np.int)
    for i in np.unique(gt):
        map[gt == i] = hex2rgb(COLOR[i])
    return map

# 计算rgb波段索引
def rgb(a_0, d):
    r = int(math.floor((700-a_0)/d))
    g = int(math.floor((546.1-a_0)/d))
    b = int(math.floor((435.8-a_0)/d))
    return r, g, b

def weight_init(m):
    if isinstance(m, nn.Linear):
        init.normal_(m.weight, 0, 5e-2)
        init.constant_(m.bias, 0)
    elif isinstance(m, nn.Conv2d):
        init.normal_(m.weight, 0, 5e-2)
        init.constant_(m.bias, 0)
    elif isinstance(m, nn.BatchNorm1d):
        init.constant_(m.weight, 1)
        init.constant_(m.bias, 0)


def loadLabel(path):
    '''
    :param path:
    :return: 训练样本标签， 测试样本标签
    '''
    assert os.path.exists(path), '{},路径不存在'.format(path)
    # keys:{train_gt, test_gt}
    gt = loadmat(path)
    return gt['train_gt'], gt['test_gt']


def splitSampleByClass(gt, ratio, seed=971104):
    '''
    :param gt: 样本标签
    :param ratio: 随机抽样每类样本的比例
    :param seed: 随机种子
    :return: 训练样本， 测试样本
    '''
    # 设置随机种子
    random.seed(seed)
    train_gt = np.zeros_like(gt)
    test_gt = np.copy(gt)
    train_indices = []
    nc = int(np.max(gt))
    # 开始随机挑选样本
    for c in range(1, nc + 1):
        samples = np.nonzero(gt == c)
        sample_indices = list(zip(*samples))
        size = int(len(sample_indices) * ratio)
        x = random.sample(sample_indices, size)
        train_indices += x
    indices = tuple(zip(*train_indices))
    train_gt[indices] = gt[indices]
    test_gt[indices] = 0
    return train_gt, test_gt


# 模型训练
def train(model, criterion, optimizer, dataLoader, device):
    '''
    :param model: 模型
    :param criterion: 目标函数
    :param optimizer: 优化器
    :param dataLoader: 批数据集
    :return: 已训练的模型，训练损失的均值
    '''
    model.train()
    model.to(device)
    trainLoss = []
    for step, (input, target) in enumerate(dataLoader):
        input, target = input.to(device), target.to(device)
        input = input.permute((0, 3, 1, 2))
        out = model(input)
        out = out.squeeze(-1).squeeze(-1) #
        loss = criterion(out, target)
        trainLoss.append(loss.item())
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if step%5 == 0:
            lr = optimizer.state_dict()['param_groups'][0]['lr']
            print('step:{} loss:{} lr:{}'.format(step, loss.item(), lr))
    return float(np.mean(trainLoss))


# 模型测试
def test(model, criterion, dataLoader, device):
    model.eval()
    model.to(device)
    evalLoss, correct = [], 0
    gt_li, pred_li = [], []
    for input, target in dataLoader:
        input, target = input.to(device), target.to(device)
        input = input.permute((0, 3, 1, 2))
        logits = model(input)
        logits = logits.squeeze(-1).squeeze(-1)
        loss = criterion(logits, target)
        evalLoss.append(loss.item())
        pred = torch.argmax(logits, dim=-1)
        correct += torch.sum(torch.eq(pred, target).int()).item()
        gt_li.extend(target.cpu().numpy())
        pred_li.extend(pred.cpu().numpy())
    acc = float(correct) / len(dataLoader.dataset)

    # 计算类别 recall 值
    class_recall = recall_score(gt_li, pred_li, average=None)
    # 计算平均 recall
    AA = class_recall.mean()
    # 计算准确率
    OA = accuracy_score(gt_li, pred_li)
    # 计算kappa
    Ka = cohen_kappa_score(gt_li, pred_li)
    return acc, np.mean(evalLoss), AA, OA, Ka

# 模型训练
def train_fusion(encoder_h, encoder_l, model_ft, criterion, optimizer, dataLoader, device):
    '''
    :param model: 模型
    :param criterion: 目标函数
    :param optimizer: 优化器
    :param dataLoader: 批数据集
    :return: 已训练的模型，训练损失的均值
    '''
    encoder_h.train()
    encoder_h.to(device)
    encoder_l.train()
    encoder_l.to(device)
    model_ft.train()
    model_ft.to(device)

    trainLoss = []
    for step, (input, target) in enumerate(dataLoader):
        input[0], input[1], target = input[0].to(device), input[1].to(device), target.to(device)
        input[0] = input[0].permute((0, 3, 1, 2))
        input[1] = input[1].permute((0, 3, 1, 2))
        # f_hsi, f_lidar = translator(input[0], input[1])
        features_hsi = encoder_h(input[0])
        features_lidar = encoder_l(input[1])
        feature1_fusion = torch.cat((features_hsi[1], features_lidar[1]), 1)
        feature2_fusion = torch.cat((features_hsi[2], features_lidar[2]), 1)
        feature3_fusion = torch.cat((features_hsi[3], features_lidar[3]), 1)
        feature4_fusion = torch.cat((features_hsi[4], features_lidar[4]), 1)

        out = model_ft(feature1_fusion, feature2_fusion, feature3_fusion, feature4_fusion)
        out = out.squeeze(-1).squeeze(-1) #
        loss = criterion(out, target)

        trainLoss.append(loss.item())
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if step%5 == 0:
            lr = optimizer.state_dict()['param_groups'][0]['lr']
            print('step:{} loss:{} lr:{}'.format(step, loss.item(), lr))
    return float(np.mean(trainLoss))

# 模型测试
def test_fusion(encoder_h, encoder_l, model_ft, criterion, dataLoader, device):
    encoder_h.eval()
    encoder_h.to(device)
    encoder_l.eval()
    encoder_l.to(device)
    model_ft.eval()
    model_ft.to(device)
    evalLoss, correct = [], 0
    gt_li, pred_li = [], []
    for input, target in dataLoader:
        input[0], input[1], target = input[0].to(device), input[1].to(device), target.to(device)
        input[0] = input[0].permute((0, 3, 1, 2))
        input[1] = input[1].permute((0, 3, 1, 2))
        # f_hsi, f_lidar = translator(input[0], input[1])
        features_hsi = encoder_h(input[0])
        features_lidar = encoder_l(input[1])
        feature1_fusion = torch.cat((features_hsi[1], features_lidar[1]), 1)
        feature2_fusion = torch.cat((features_hsi[2], features_lidar[2]), 1)
        feature3_fusion = torch.cat((features_hsi[3], features_lidar[3]), 1)
        feature4_fusion = torch.cat((features_hsi[4], features_lidar[4]), 1)

        logits = model_ft(feature1_fusion, feature2_fusion, feature3_fusion, feature4_fusion)
        logits = logits.squeeze(-1).squeeze(-1)
        loss = criterion(logits, target)
        evalLoss.append(loss.item())
        pred = torch.argmax(logits, dim=-1)
        correct += torch.sum(torch.eq(pred, target).int()).item()
        gt_li.extend(target.cpu().numpy())
        pred_li.extend(pred.cpu().numpy())
    acc = float(correct) / len(dataLoader.dataset)

    # 计算类别 recall 值
    class_recall = recall_score(gt_li, pred_li, average=None)
    # 计算平均 recall
    AA = class_recall.mean()
    # 计算准确率
    OA = accuracy_score(gt_li, pred_li)
    # 计算kappa
    Ka = cohen_kappa_score(gt_li, pred_li)
    return acc, np.mean(evalLoss), AA, OA, Ka
    # return acc, np.mean(evalLoss)
